unit hiFramedImage;

// DsFramedImage Release 1.1 for Delphi 3/4
// Copyright (c) 1999-2000 by Djoko Susilo
// --------- Q14 Project Nov-99 --------------
// e-mail: djokos@cabi.net.id
// http://homepage.indo1.com/delphianhome/
// -------------------------------------------

interface

uses
  Kol, Windows, share, debug, Win, Messages, ActiveX;

type
  TFrameCorner = (fcSquare, fcRound);
  TFrameStyle = (fsNone, fsGradient, fsFlat, fsGradFlat, fsFlatGrad,
                 fsInGrad, fsOutGrad);
  THIFramedImage = class(THIWin)
  private
    FPicture: Kol.PBitmap;
    FColor_High: TColor;
    FColor_Low: TColor;
    FFrameColor: TColor;
    FFrameCorner: TFrameCorner;
    FFrameStyle: TFrameStyle;
    FFrameWidth: Integer;
    FCornerRadius: Integer;
    FShadowColor: TColor;
    F, RgnF: HRgn;
    FActualSize: Boolean;
    procedure _onResize(Obj:PObj);
  protected
    procedure _OnClick(Obj:PObj);
    procedure _OnPaint(Sender: PControl; DC: HDC);
    procedure SetPicture(Value: Kol.PBitmap);
    procedure SetBitmap(Value: HBitmap);
    procedure SetActualSize(Value: Boolean);
    procedure SetColor_High(Value: TColor);
    procedure SetColor_Low(Value: TColor);
    procedure SetCornerRadius(Value: Integer);
    procedure SetFrameColor(Value: TColor);
    procedure SetFrameCorner(Value: TFrameCorner);
    procedure SetFrameStyle(Value: TFrameStyle);
    procedure SetFrameWidth(Value: Integer);
    procedure SetShadowColor(Value: TColor);
    procedure GradientOuter(Rd: Integer);
    procedure GradientInner(Rd: Integer);
    procedure DrawPicture;
  public
    _prop_FrameStyle: byte;
    _prop_FrameCorner: byte;
    _prop_FileName: string;
    _prop_Image: PStream;

    _data_Stream: THI_Event;
    _data_FileName: THI_Event;
    _event_onClick: THI_Event;

    procedure Init; override;
    destructor Destroy; override;

    procedure _work_doLoadBitmap(var _Data:TData; Index:word);
    procedure _work_doLoadFileImage(var _Data:TData; Index:word);
    procedure _work_doLoadFromStream(var _Data:TData; Index:word);
    procedure _work_doFrameStyle(var _Data:TData; Index:word);
    procedure _work_doFrameCorner(var _Data:TData; Index:word);

    property _prop_Bitmap:HBITMAP write SetBitmap;
    property _prop_ActualSize: Boolean read FActualSize write SetActualSize;
    property _prop_ColorHigh: TColor read FColor_High write SetColor_High;
    property _prop_ColorLow: TColor read FColor_Low write SetColor_Low;
    property _prop_FrameColor: TColor read FFrameColor write SetFrameColor;
    property _prop_FrameWidth: Integer read FFrameWidth write SetFrameWidth;
    property _prop_CornerRadius: Integer read FCornerRadius write SetCornerRadius;
    property _prop_ShadowColor: TColor read FShadowColor write SetShadowColor;

    property Bitmap: Kol.PBitmap read FPicture write SetPicture;
  end;


implementation

const
  GDIP = 'gdiplus.dll';

var PicRect: TRect; {--Bitmap area and position--}
    x0, x1, y0, y1: Integer;  {--frame area--}

    GdiplusToken: ULONG = 0;

type
  TGdiplusStartupInput = packed record
    GdiplusVersion          : Cardinal;       // Must be 1
    DebugEventCallback      : Pointer;        // Ignored on free builds
    SuppressBackgroundThread: BOOL;           // FALSE unless you're prepared to call
                                              // the hook/unhook functions properly
    SuppressExternalCodecs  : BOOL;           // FALSE unless you want GDI+ only to use
  end;

  TStatus = (
    Ok,
    GenericError,
    InvalidParameter,
    OutOfMemory,
    ObjectBusy,
    InsufficientBuffer,
    NotImplemented,
    Win32Error,
    WrongState,
    Aborted,
    FileNotFound,
    ValueOverflow,
    AccessDenied,
    UnknownImageFormat,
    FontFamilyNotFound,
    FontStyleNotFound,
    NotTrueTypeFont,
    UnsupportedGdiPlusVersion,
    GdiplusNotInitialized,
    PropertyNotFound,
    PropertyNotSupported,
    ProfileNotFound
  );

  ARGB = Cardinal;

  TGpBitmap = record
    Handle: Pointer;
  end;

  function GdiplusStartup(out token: ULONG; const input: TGdiplusStartupInput; output: Pointer): TStatus; stdcall; external GDIP;
  function GdipCreateBitmapFromStream(stream: IStream; out image: TGpBitmap): TStatus; stdcall; external GDIP;
  function GdipCreateHBITMAPFromBitmap(const Bitmap: TGpBitmap; var Handle: HBitmap; Background: ARGB): TStatus; stdcall; external GDIP;
  function GdipDisposeImage(var Image: TGpBitmap): TStatus; stdcall; external GDIP;



function InitGDIPlus: Boolean;
var
  StartupInput: TGdiplusStartupInput;
begin
  if GdiplusToken = 0 then
  begin
    FillChar(StartupInput, SizeOf(StartupInput), 0);
    StartupInput.GdiplusVersion := 1;
    Result := GdiplusStartup(GdiplusToken, StartupInput, nil) = Ok;
  end else begin
    Result := True;
  end;
end;

function GDIPLoadBitmapFromStream(Bitmap: Kol.PBitmap; Stream: PStream): Boolean;
var
  IStr: iStream;
  HMemo: HGLOBAL;
  Buf: Longint;
  L: Largeint;
  GpBitmap: TGpBitmap;
  HBmp: HBitmap;
begin
  Result := False;
  Stream.Position := 0;
  HMemo:=GlobalAlloc(GMEM_FIXED,Stream.Size);
  if CreateStreamOnHGlobal(HMemo,True,IStr) = S_OK then
  begin
    IStr.write(Stream.Memory,Stream.Size,@Buf);
    IStr.Seek(0,STREAM_SEEK_SET,L);
  end else Exit;

  if not InitGDIPlus then
    Exit;
  if GdipCreateBitmapFromStream(IStr, GpBitmap) = Ok then
  begin
    if GdipCreateHBITMAPFromBitmap(GpBitmap, HBmp, $FFFFFFFF) = Ok then
    begin
      Bitmap.Handle := HBmp;
      Result := True;
    end;
    GdipDisposeImage(GpBitmap);
  end;
end;

procedure THIFramedImage._work_doLoadBitmap;
begin
  FPicture := ToBitmap(_Data);
  Control.Invalidate;
end;

procedure THIFramedImage._work_doLoadFileImage;
var
  Ps,Fs: PStream;
  fn: string;
begin
  fn := ReadString(_Data,_data_FileName,_prop_FileName);
  if fn = '' then exit;
  if not FileExists(fn) then exit;
  Ps := NewMemoryStream;
  Fs := NewReadFileStream(fn);
  Ps.Position := 0;
  Fs.Position := 0;
  Stream2Stream(Ps,Fs,Fs.Size);
  if Ps <> nil then
  begin
    {FPicture.PixelFormat := pf32bit;
    FPicture.Canvas.Brush.Color := Color2RGB(Control.Color);
    FPicture.Canvas.Brush.BrushStyle := bsSolid;
    FPicture.Canvas.FillRect(Kol.MakeRect(0,0,Control.Width,Control.Height));
    }
    GDIPLoadBitmapFromStream(FPicture,ps);
  end else Exit;
  Control.Invalidate;
  Fs.Free;
  Ps.Free;
end;

procedure THIFramedImage._work_doLoadFromStream;
var
  ps: PStream;
begin
  ps := ReadStream(_Data, _data_Stream,nil);
  if ps <> nil then
  begin
    {FPicture.PixelFormat := pf32bit;
    FPicture.Canvas.Brush.Color := Color2RGB(Control.Color);
    FPicture.Canvas.Brush.BrushStyle := bsSolid;
    FPicture.Canvas.FillRect(MakeRect(0,0,Control.Width,Control.Height));
    }
    GDIPLoadBitmapFromStream(FPicture,ps);
  end;
  Control.Invalidate;
end;

procedure THIFramedImage._work_doFrameStyle;
begin
  FFrameStyle := TFrameStyle(ToInteger(_Data));
  Control.Invalidate;
end;

procedure THIFramedImage._work_doFrameCorner;
begin
  FFrameCorner := TFrameCorner(ToInteger(_Data));
  Control.Invalidate;
end;

procedure THIFramedImage._OnClick(Obj:PObj);
begin
  _hi_onEvent(_event_OnClick);
end;

procedure THIFramedImage.Init;
begin
  //Control := _NewControl(FParent,'THIFramedImage', WS_VISIBLE or WS_CHILD,False,nil);
  Control := NewPaintbox(FParent);
  inherited;
  Control.SubClassName := 'THIFramedImage';
  FFrameColor :=_prop_Color;
  FFrameCorner := TFrameCorner(_prop_FrameCorner);
  FFrameStyle := TFrameStyle(_prop_FrameStyle);
  if _prop_Image <> nil then
    GDIPLoadBitmapFromStream(FPicture,_prop_Image);
  Control.OnPaint := _OnPaint;
  Control.OnClick := _OnClick;
end;

destructor THIFramedImage.Destroy;
begin
  FPicture.Free;
  inherited Destroy;
end;

procedure THIFramedImage.GradientOuter(Rd: Integer);
var
  FC, Warna: TColor;
  R, G, B: Byte;
  AwalR, AwalG, AwalB,
  AkhirR, AkhirG, AkhirB,
  y, t: Integer;
begin
  if (FFrameStyle=fsInGrad) or (FFrameStyle=fsOutGrad) then t:=FFrameWidth
  else if FFrameWidth mod 2=0 then t:=FFrameWidth div 2
  else t:=(FFrameWidth+1) div 2;
  Warna:=Color2RGB(FFrameColor);
  FC:=Color2RGB(FShadowColor);

  AwalR:=GetRValue(FC); AkhirR:=GetRValue(Warna);
  AwalG:=GetGValue(FC); AkhirG:=GetGValue(Warna);
  AwalB:=GetBValue(FC); AkhirB:=GetBValue(Warna);

  for y:=0 to t-1 do
  begin
    R:=AwalR+Round(Sqrt(t*t-Sqr(t-y))*(AkhirR-AwalR)/t);
    G:=AwalG+Round(Sqrt(t*t-Sqr(t-y))*(AkhirG-AwalG)/t);
    B:=AwalB+Round(Sqrt(t*t-Sqr(t-y))*(AkhirB-AwalB)/t);
    Control.Canvas.Brush.Color:=RGB(R, G, B);
    if (Rd-y)>0 then
      RgnF:=CreateRoundRectRgn(x0+y, y0+y, x1+1-y, y1+1-y, 2*(Rd-y), 2*(Rd-y))
    else RgnF:=CreateRectRgn(x0+y, y0+y, x1-y, y1-y);
    try
      FrameRgn(Control.Canvas.Handle, RgnF, Control.Canvas.Brush.Handle, 1, 1);
    finally
      DeleteObject(RgnF);
    end;
  end;
end;

procedure THIFramedImage.GradientInner(Rd: Integer);
var
  FC, Warna: TColor;
  R, G, B: Byte;
  AwalR, AwalG, AwalB,
  AkhirR, AkhirG, AkhirB,
  y, t, n: Integer;
begin
  if (FFrameStyle=fsInGrad) or (FFrameStyle=fsOutGrad) then
    if FFrameWidth mod 2=0 then t:=FFrameWidth else t:=FFrameWidth+1
  else if FFrameWidth mod 2=0 then t:=FFrameWidth div 2
  else t:=(FFrameWidth+1) div 2;

  if FFrameStyle=fsInGrad then n:=0 else n:=t;

  Warna:=Color2RGB(FFrameColor);
  FC:=Color2RGB(FShadowColor);

  AwalR:=GetRValue(FC); AkhirR:=GetRValue(Warna);
  AwalG:=GetGValue(FC); AkhirG:=GetGValue(Warna);
  AwalB:=GetBValue(FC); AkhirB:=GetBValue(Warna);

  for y:=t downto 0 do
  begin
    R:=AwalR+Round(Sqrt(t*t-y*y)*(AkhirR-AwalR)/t);
    G:=AwalG+Round(Sqrt(t*t-y*y)*(AkhirG-AwalG)/t);
    B:=AwalB+Round(Sqrt(t*t-y*y)*(AkhirB-AwalB)/t);
    Control.Canvas.Brush.Color:=RGB(R, G, B);
    if Rd-t-y>0 then
      RgnF:=CreateRoundRectRgn(x0+n+y-1, y0+n+y-1, x1-n-y+2, y1-n-y+2,
                               2*(Rd-y-t), 2*(Rd-y-t))
    else RgnF:=CreateRectRgn(x0+n+y-1, y0+n+y-1, x1-n-y+1, y1-n-y+1);
    try
      FrameRgn(Control.Canvas.Handle, RgnF, Control.Canvas.Brush.Handle, 1, 1);
    finally
      DeleteObject(RgnF);
    end;
  end;
end;

procedure THIFramedImage._OnPaint(Sender: PControl; DC: HDC);
var Rd, t: Integer;
begin
  {--check frame width and corner radius--}
  t:=FFrameWidth;
  if t mod 2<>0 then Inc(t);
  Rd:=2*FCornerRadius;

  {--display Bitmap--}
  DrawPicture;

  {--draw component border to identify current position--}
  //       desing-time :)
  //if csDesigning in ComponentState then
  {begin
    Control.Canvas.Brush.BrushStyle:=bsClear;
    Control.Canvas.Pen.Color:=clBtnShadow;
    Control.Canvas.Pen.PenStyle:=psDash;
    Control.Canvas.Rectangle(0, 0, Control.Width, Control.Height);
  end;}

  {--create and fill frame area--}
  if FFrameStyle=fsNone then Exit;
  if Bitmap=nil then
  begin
    x0:=0;      y0:=0;
    x1:=Control.Width;  y1:=Control.Height;
  end
  else if FActualSize then
  case _prop_Align of
    caLeft, caRight:
      begin
        x0:=0;      y0:=PicRect.Top-t;
        x1:=Control.Width;  y1:=PicRect.Bottom+t;
      end;
    caTop, caBottom:
      begin
        x0:=PicRect.Left-t;  y0:=0;
        x1:=PicRect.Right+t; y1:=Control.Height;
      end;
    caNone:
      begin
        x0:=0;      y0:=0;
        x1:=Control.Width;  y1:=Control.Height;
      end;
    caClient:
      begin
        x0:=PicRect.Left-t;
        x1:=PicRect.Right+t;
        y0:=PicRect.Top-t;
        y1:=PicRect.Bottom+t;
      end;
  end
  else
  begin
    x0:=PicRect.Left-t;   y0:=PicRect.Top-t;
    x1:=PicRect.Right+t;  y1:=PicRect.Bottom+t;
  end;

  if (FFrameStyle<>fsInGrad) and (FFrameStyle<>fsOutGrad) then
  begin
    {--outer region--}
    if (FFrameCorner=fcSquare) or (FCornerRadius=0) then
      F:=CreateRectRgn(x0, y0, x1-2, y1-2)
    else F:=CreateRoundRectRgn(x0, y0, x1-1, y1-1, Rd, Rd);
    {--inner region--}
    if (Rd-2*t)<1 then RgnF:=CreateRectRgn(x0+t-2, y0+t-2, x1-t, y1-t)
    else RgnF:=CreateRoundRectRgn(x0+t-2, y0+t-2, x1-t+1, y1-t+1, Rd-2*t, Rd-2*t);
    try
      CombineRgn(RgnF, RgnF, F, RGN_XOR);
      {--make 3D effect--}
      Control.Canvas.Brush.Color:=FColor_High;
      FillRgn(Control.Canvas.Handle, RgnF, Control.Canvas.Brush.Handle);
      OffsetRgn(RgnF, 2, 2);
      Control.Canvas.Brush.Color:=FColor_Low;
      FillRgn(Control.Canvas.Handle, RgnF, Control.Canvas.Brush.Handle);
      OffsetRgn(RgnF, -1, -1);
      Control.Canvas.Brush.Color:=FFrameColor;
      FillRgn(Control.Canvas.Handle, RgnF, Control.Canvas.Brush.Handle);
    finally
      {--freed the memory used by regions--}
      DeleteObject(F);
      DeleteObject(RgnF);
    end;
  end;
      
  case FFrameStyle of
    fsGradient:
       begin
         if FFrameCorner=fcRound then GradientOuter(FCornerRadius)
         else GradientOuter(0);
         GradientInner(FCornerRadius);
       end;
    fsGradFlat,  {--outer: gradient, inner: flat--}
    fsOutGrad:   {--outer: gradient, inner: none--}
       begin
         if FFrameCorner=fcRound then GradientOuter(FCornerRadius)
         else GradientOuter(0);
       end;
    fsFlatGrad:  {--outer: flat, inner: gradient--}
       begin
         GradientInner(FCornerRadius);
       end;
    fsInGrad:    {--outer: none, inner: gradient--}
       begin
         if FFrameCorner=fcSquare then GradientInner(0)
         else GradientInner(FCornerRadius+t);
       end;
  end; {case}
end;

procedure THIFramedImage.DrawPicture;
var x, y: Integer;
    PicX, PicY: Integer;
    x1, y1: real;
    t: Integer;

begin
  t:=FFrameWidth;
  if t mod 2<>0 then Inc(t);
  if FFrameStyle=fsNone then t:=0;

  Control.Canvas.Brush.Color := Color2RGB(Control.Color);
  Control.Canvas.Brush.BrushStyle := bsSolid;
  Control.Canvas.FillRect(MakeRect(0,0,Control.Width,Control.Height));

  if Bitmap<>nil then
  begin
    if FActualSize then
    begin
      case _prop_Align of {--check align value--}
        caTop, caBottom:
          begin
            Control.Height:=Bitmap.Height+2*t;
            PicX:=(Control.Width-Bitmap.Width) div 2;
            PicY:=t;
            PicRect.Top:=0+t;
            PicRect.Bottom:=Control.Height-t;
            PicRect.Left:=PicX;
            PicRect.Right:=PicRect.Left+Bitmap.Width;
            //Canvas.Draw(PicX, PicY, Bitmap);
         end;
        caLeft, caRight:
          begin
            Control.Width:=Bitmap.Width+2*t;
            PicX:=t;
            PicY:=(Control.Height-Bitmap.Height) div 2;
            PicRect.Top:=PicY;
            PicRect.Bottom:=PicRect.Top+Bitmap.Height;
            PicRect.Left:=0+t;
            PicRect.Right:=Control.Width-t;
            //Canvas.Draw(PicX, PicY, Bitmap);
          end;
        caClient:
          begin
            PicX:=(Control.Width-Bitmap.Width) div 2;
            PicY:=(Control.Height-Bitmap.Height) div 2;
            PicRect.Left:=PicX;
            PicRect.Right:=PicRect.Left+Bitmap.Width;
            PicRect.Top:=PicY;
            PicRect.Bottom:=PicRect.Top+Bitmap.Height;
            //Canvas.Draw(PicX, PicY, Bitmap);
          end;
        caNone:
          begin
            Control.Width:=Bitmap.Width+2*t;
            Control.Height:=Bitmap.Height+2*t;
            PicX:=t;
            PicY:=t;
            //Canvas.Draw(PicX, PicY, Bitmap);
          end;
      end;
      //Control.Canvas.Draw(PicX, PicY, Bitmap);
      Bitmap.Draw(Control.Canvas.Handle,PicX, PicY);
    end
    else
    begin
      x1:=(Control.Width-2*t)/Bitmap.Width;
      y1:=(Control.Height-2*t)/Bitmap.Height;
      if x1>y1 then  {--fit to height--}
      begin
        PicRect.Top:=0+t;
        PicRect.Bottom:=Control.Height-t;
        x:=Round(Bitmap.Width*y1);
        PicRect.Left:=(Control.Width-x) div 2;
        PicRect.Right:=PicRect.Left+x;
      end
      else  {--fit to width--}
      begin
        PicRect.Left:=0+t;
        PicRect.Right:=Control.Width-t;
        y:=Round(Bitmap.Height*x1);
        PicRect.Top:=(Control.Height-y)div 2;
        PicRect.Bottom:=PicRect.Top+y;
      end;
      //Control.Canvas.StretchDraw(PicRect, Bitmap);
      Bitmap.StretchDraw(Control.Canvas.Handle,PicRect);
    end;
  end;
end;

procedure THIFramedImage.SetFrameCorner(Value: TFrameCorner);
begin
  if Value<>FFrameCorner then
    FFrameCorner:=Value;
end;

procedure THIFramedImage.SetFrameStyle(Value: TFrameStyle);
begin
  if Value<>FFrameStyle then
    FFrameStyle:=Value;
end;

procedure THIFramedImage.SetColor_High(Value: TColor);
begin
  if Value<>FColor_High then
    FColor_High:=Value;
end;

procedure THIFramedImage.SetColor_Low(Value: TColor);
begin
  if Value<>FColor_Low then
    FColor_Low:=Value;
end;

procedure THIFramedImage.SetFrameColor(Value: TColor);
begin
  if Value<>FFrameColor then
    FFrameColor:=Value;
end;

procedure THIFramedImage.SetShadowColor(Value: TColor);
begin
  if Value<>FShadowColor then
    FShadowColor:=Value;
end;

procedure THIFramedImage.SetCornerRadius(Value: Integer);
begin
  if Value<>FCornerRadius then
  begin
    if Value<0 then FCornerRadius:=0
    else if Value>30 then FCornerRadius:=30
    else FCornerRadius:=Value;
    if FCornerRadius>4*FFrameWidth then FCornerradius:=4*FFrameWidth-1;
    //Control.Invalidate;
  end;
end;

procedure THIFramedImage.SetFrameWidth(Value: Integer);
begin
  if Value<>FFrameWidth then
  begin
    if Value<4 then FFrameWidth:=4         {--minimum value--}
    else if Value>20 then FFrameWidth:=20  {--maximum value--}
    else FFrameWidth:=Value;
    if FCornerRadius>4*FFrameWidth then FCornerradius:=4*FFrameWidth-1;
    //Control.Invalidate;
  end;
end;

procedure THIFramedImage.SetBitmap(Value: HBitmap);
begin
  if Value <> 0 then
  begin
    if not Assigned(FPicture) then
      FPicture:=NewBitmap(1,1);
    FPicture.Handle := Value;
  end;
end;

procedure THIFramedImage.SetPicture(Value: Kol.PBitmap);
begin
  if Value = nil then Exit;
  if not Assigned(FPicture) then
    FPicture:=NewBitmap(1,1);
  FPicture.Assign(Value);
end;


procedure THIFramedImage.SetActualSize(Value: Boolean);
begin
  if (Value<>FActualSize) then
    FActualSize:=Value;
end;

procedure THIFramedImage._onResize(Obj:PObj);
var x: Integer;
begin
  inherited;
  if Control.Width>Control.Height then x:=Control.Height div 2
  else x:=Control.Width div 2;
  if FFrameWidth>x then
    begin FFrameWidth:=x; Control.Invalidate; end;
end;

end.
